// ==========  This file is under  LGPL, the GNU Lesser General Public Licence
// ==========  Dialing Lemmatizer (www.aot.ru)
// ==========  Copyright by Alexey Sokirko
// ==========  Changed by Pavel Makarov (January 2007), e-mail: PavMakarov@gmail.com

#include "stdafx.h"

#include "StdMorph.h"
#include "util_classes.h"
#include "Lemmatizers.h"
#include "utilit.h"

CLemmatizer::CLemmatizer(MorphLanguageEnum Language) : CMorphDict(Language)
{	
	m_bLoaded = false;
	m_bAllowRussianJo = false;

	//      
	InitAutomat(new CMorphAutomat(Language, MorphAnnotChar));
}

CLemmatizer::~CLemmatizer()
{
}

// CLemmatizer::LemmatizeWord should return true if the word was found in the dictionary
bool CLemmatizer::LemmatizeWord(string& InputWordStr, vector<CAutomAnnotationInner>& results) const
{
	//      
	RmlMakeUpper(InputWordStr, GetLanguage());

	size_t WordOffset = 0;	//     
	
	//       
	// (  ,    ,  )
	//        
	m_pFormAutomat->GetInnerMorphInfos(InputWordStr, 0, results);

	return !results.empty();
}

//////////////////////////////////////////////////////////////////////////////
//     morph.options
void CLemmatizer::ReadOptions(wstring wsFileName)
{
	string Options;
	LoadFileToString(wsFileName, Options);
	StringTokenizer lines(Options.c_str(), "\r\n");
	while (lines())
	{
		string line = lines.val();
		Trim(line);
		if (line.empty()) continue;

		//   AllowRussianJo (  )
		//     - ""  
		if (line == "AllowRussianJo")
			m_bAllowRussianJo = true;
	}
}

//    
bool CLemmatizer::LoadDictionariesRegistry(wstring& wsDictPath, string& strError)
{
	try
	{
		m_bLoaded = Load(wsDictPath + MORPH_MAIN_FILES);	//    

		if (!m_bLoaded) 
		{
			wstring wsError(L"Cannot load " + wsDictPath + MORPH_MAIN_FILES);
			ConvertWStrToStrSysLoc(wsError, strError);			
			return false;
		}

		ReadOptions(wsDictPath + OPTIONS_FILE);	//     morph.options

		//    
		m_PrefixesSet.clear();
		m_PrefixesSet.insert(m_Prefixes.begin(), m_Prefixes.end() );

		return m_bLoaded;
	}
	catch(CExpc& s)
	{
		strError = s.m_strCause;
		return false;
	}
	catch(...)
	{
		strError = "General exception";
		return false;
	}
}

CLemmatizerRussian::CLemmatizerRussian() : CLemmatizer(morphRussian)
{
}

void CLemmatizerRussian::FilterSrc(string& src) const	
{
	if (!m_bAllowRussianJo)
		ConvertJO2Je(src); 
}

CLemmatizerEnglish:: CLemmatizerEnglish() : CLemmatizer(morphEnglish)
{
}

void CLemmatizerEnglish:: FilterSrc(string& src) const
{
}

CLemmatizerGerman:: CLemmatizerGerman() : CLemmatizer(morphGerman)
{
};

void CLemmatizerGerman:: FilterSrc(string& src) const
{
};

//   -   
CLemmatizerRussian*	gRusLemmatizer = NULL;
CLemmatizerEnglish*	gEngLemmatizer = NULL;
CLemmatizerGerman*	gGerLemmatizer = NULL;

bool	LoadLemmatizer(wchar_t* wcDictPath, int nLang, double* pdLoadTime)
{
	if (IsLemmatizerLoad(nLang))
	{
		return true;
	}

	CLemmatizer* pLemmatizer;

	switch (nLang)
	{
	case morphRussian:
		gRusLemmatizer = new CLemmatizerRussian;
		pLemmatizer	= gRusLemmatizer;
		break;

	case morphEnglish:
		gEngLemmatizer = new CLemmatizerEnglish;
		pLemmatizer	= gEngLemmatizer;
		break;

	case morphGerman:
		gGerLemmatizer = new CLemmatizerGerman;
		pLemmatizer	= gGerLemmatizer;
		break;

	default:
		#ifndef _SILENT
			::MessageBox(NULL, "This language is not support", "Error", MB_ICONERROR);
		#endif
		return false;
	}

	LARGE_INTEGER liFreq;
	LARGE_INTEGER liCountFirst, liCountSecond;

	QueryPerformanceFrequency(&liFreq);
	liCountFirst.QuadPart = liCountSecond.QuadPart = 0;

	QueryPerformanceCounter(&liCountFirst);

	// loading dictionary into the memory
	string strError;
	wstring wsDictPath;

	wsDictPath.assign(wcDictPath);

	if (!pLemmatizer->LoadDictionariesRegistry(wsDictPath, strError))
	{ 
		UnloadLemmatizer(nLang);
		return false;
	}	

	QueryPerformanceCounter(&liCountSecond);

	//    
	if (pdLoadTime != NULL)
	{
		*pdLoadTime = (double) (liCountSecond.QuadPart - liCountFirst.QuadPart) / (double) liFreq.QuadPart;
	}

	return true;
}

bool	IsLemmatizerLoad(int nLang)
{
	switch (nLang)
	{
	case morphRussian:
		return gRusLemmatizer != NULL;
		break;

	case morphEnglish:
		return gEngLemmatizer != NULL;
		break;

	case morphGerman:
		return gGerLemmatizer != NULL;
		break;
	}

	return false;
}

void	UnloadLemmatizer(int nLang)
{
	if (IsLemmatizerLoad(nLang))
	{
		switch (nLang)
		{
		case morphRussian:
			delete gRusLemmatizer;
			gRusLemmatizer = NULL;
			break;

		case morphEnglish:
			delete gEngLemmatizer;
			gEngLemmatizer = NULL;
			break;

		case morphGerman:
			delete gGerLemmatizer;
			gGerLemmatizer = NULL;
			break;
		}
	}
}

//     (  )
bool	GetLemmas(char* cInputWord, int nLang, char* cLemmas, int nMaxBufferSize, char cLemmasSeparator)
{
	strcpy(cLemmas, "");	//     

	if (!IsLemmatizerLoad(nLang))	//   
	{
		return false;
	}

	CLemmatizer* pLemmatizer;

	switch (nLang)
	{
	case morphRussian:
		pLemmatizer	= gRusLemmatizer;
		break;

	case morphEnglish:
		pLemmatizer	= gEngLemmatizer;
		break;

	case morphGerman:
		pLemmatizer	= gGerLemmatizer;
		break;

	default:
		return false;
	}

	string InputWordStr;
	InputWordStr.assign(cInputWord);	//   char*  string

	//    (  :  ""  "")
	pLemmatizer->FilterSrc(InputWordStr);

	//     .  : 
	//   ,     ,      
	vector<CAutomAnnotationInner>	FindResults;

	//   
	bool bFound = pLemmatizer->LemmatizeWord(InputWordStr, FindResults);

	size_t	Count = FindResults.size();	//    
	size_t	OutLen = 0;	//   

	for (DWORD i = 0; i < Count; i++)	//      
	{
		//    
		const CAutomAnnotationInner& A = FindResults[i];

		//  :          	
		const CFlexiaModel& M = pLemmatizer->m_FlexiaModels[A.m_ModelNo];

		//   : , ,   
		const CMorphForm& F = M.m_Flexia[A.m_ItemNo];

		size_t PrefixLen = F.m_PrefixStr.length();	//    
		size_t BaseStart  = 0;

		//     
		if (PrefixLen > 0 && !strncmp(InputWordStr.c_str(), F.m_PrefixStr.c_str(), PrefixLen))
		{
			BaseStart = PrefixLen;
		}

		//    
		int BaseLen  = InputWordStr.length() - F.m_FlexiaStr.length() - BaseStart;

		if (BaseLen < 0)
		{
			BaseLen = InputWordStr.length();
		}

		//      
		size_t FlexiaLength = M.m_Flexia[0].m_FlexiaStr.length();

		//     
		if (BaseLen + FlexiaLength + 2 > nMaxBufferSize - OutLen) // + 2 => + cLemmasSeparator + '\0'
		{
			return false; 
		}

		//   (    )
		strncpy(cLemmas + OutLen, InputWordStr.c_str() + BaseStart, BaseLen);
		OutLen += BaseLen;

		//   . . .  =  +  
		strncpy(cLemmas + OutLen, M.m_Flexia[0].m_FlexiaStr.c_str(), FlexiaLength);
		OutLen += FlexiaLength;

		//  ,    
		cLemmas[OutLen] = cLemmasSeparator;
		OutLen++;
	}

	cLemmas[OutLen] = '\0';	//   

	return bFound;
}

