// ==========  This file is under  LGPL, the GNU Lesser General Public Licence
// ==========  Dialing Lemmatizer (www.aot.ru), 
// ==========  Copyright by Alexey Sokirko (2004)
// ==========  Changed by Pavel Makarov (January 2007), e-mail: PavMakarov@gmail.com

#include "stdafx.h"
#include "StdMorph.h"
#include "MorphDict.h"

//======================================================
//=============		CMorphDict			   =============
//======================================================

CMorphDict::CMorphDict(MorphLanguageEnum Language) : 
	m_SearchInfoLess(m_Bases)
{
	m_pFormAutomat  = 0;
}

CMorphDict::~CMorphDict() 
{
	if (m_pFormAutomat != 0)
		delete m_pFormAutomat;
	m_pFormAutomat = 0;
}

void CMorphDict::InitAutomat(CMorphAutomat*	pFormAutomat)
{
	assert (m_pFormAutomat == 0);
	assert (pFormAutomat != 0);
	m_pFormAutomat = pFormAutomat;
}

void	CMorphDict::GetLemmaInfos (const string& Text, size_t TextPos, vector<CAutomAnnotationInner>& Infos) const 
{
	const size_t TextLength = Text.length();
	size_t Count = 	Infos.size();
	
	for (size_t i=0; i<Count; i++)
	{
		//  :   ,    ,
		//  ,      
		const CAutomAnnotationInner&  A = Infos[i];

		//  : ,   : , , 
		const CFlexiaModel& F = m_FlexiaModels[A.m_ModelNo];

		//   : , ,  
		const CMorphForm& M = F.m_Flexia[A.m_ItemNo];

		//     
		size_t TextStartPos = TextPos + m_Prefixes[A.m_PrefixNo].length() + M.m_PrefixStr.length();

		//    (   )
		string Base = m_Prefixes[A.m_PrefixNo]+Text.substr(TextStartPos, TextLength-TextStartPos-M.m_FlexiaStr.length());

		//    (,     )
		vector<CLemmaInfoAndLemma>::const_iterator start = m_LemmaInfos.begin()+m_ModelsIndex[A.m_ModelNo];
		vector<CLemmaInfoAndLemma>::const_iterator end = m_LemmaInfos.begin()+m_ModelsIndex[A.m_ModelNo+1];

		//  ,    ,   :
		//          ,
		//   ,     
		vector<CLemmaInfoAndLemma>::const_iterator it =  
			lower_bound (start, end, Base.c_str(), m_SearchInfoLess);
	
		assert (it != m_LemmaInfos.end());
		{
			int LemmaStrNo = it->m_LemmaStrNo;

			// :   
			assert (Base == m_Bases[LemmaStrNo].GetString());
		}

		//     
		Infos[i].m_LemmaInfoNo = it-m_LemmaInfos.begin();
	}
}

inline size_t get_size_in_bytes (const CLemmaInfoAndLemma& t)
{
	return		get_size_in_bytes(t.m_LemmaInfo) 
		+	get_size_in_bytes(t.m_LemmaStrNo);
}

inline size_t restore_from_bytes(CLemmaInfoAndLemma& t, const BYTE* buf)
{
	buf += restore_from_bytes(t.m_LemmaInfo, buf);
	buf += restore_from_bytes(t.m_LemmaStrNo, buf);
	return get_size_in_bytes(t);
}

// This procedure builds CMorphDict::m_ModelsIndex, which is an index to CMorphDict::m_LemmaInfos
// *  m_LemmaInfos is sorted by m_LemmaInfo.m_FlexiaModelNo
// *  CMorphDict::m_ModelsIndex stores all periods of  CMorphDict::m_ModelsIndex with equal  m_LemmaInfo.m_FlexiaModelNo
// *  if a=m_ModelsIndex[i] and b=m_ModelsIndex[i+1], then for each j (a<=j<b)
//     LemmaInfos[j].m_LemmaInfo.m_FlexiaModelNo == i
// *  for some i m_ModelsIndex[i] can be equal to m_ModelsIndex[i+1], it means
//    that flexia model i is not used. To delete unused models the dictionary should be packed.
//
// CMorphDict::m_ModelsIndex - ,   CMorphDict::m_LemmaInfos,
//   m_LemmaInfo.m_FlexiaModelNo.
// CMorphDict::m_ModelsIndex    CMorphDict::m_ModelsIndex
//   m_LemmaInfo.m_FlexiaModelNo.  a=m_ModelsIndex[i]  b=m_ModelsIndex[i+1],
//    j (a<=j<b): LemmaInfos[j].m_LemmaInfo.m_FlexiaModelNo == i.
// :     ,     ,
//    .
void CMorphDict::CreateModelsIndex()
{
	size_t i;

	m_ModelsIndex.clear();
	if (m_LemmaInfos.empty()) return;

	//    1 ,    
	m_ModelsIndex.resize(m_FlexiaModels.size() + 1);

	//   
	int CurrentModel  = m_LemmaInfos[0].m_LemmaInfo.m_FlexiaModelNo; 
	m_ModelsIndex[CurrentModel] = 0;

	//     
	for (i = 0; i < m_LemmaInfos.size(); i++)
	{
		// ,             
		//      m_LemmaInfos      
		for (; CurrentModel < m_LemmaInfos[i].m_LemmaInfo.m_FlexiaModelNo; CurrentModel++ )
		{
//			int debug = m_LemmaInfos[i].m_LemmaInfo.m_FlexiaModelNo; //  

			//           1     
			//  ,          
			// :     (n)    ,  
			//  .   (m)    :
			// m_ModelsIndex[n] <= m < m_ModelsIndex[n + 1]
			m_ModelsIndex[CurrentModel + 1] = i;
		}
	}
	
	//     
	for (; CurrentModel < m_FlexiaModels.size(); CurrentModel++ )
		m_ModelsIndex[CurrentModel + 1] = m_LemmaInfos.size();

//	//   ( )
//	for (i = 0; i < m_LemmaInfos.size(); i++)
//	{
//		int debug = m_LemmaInfos[i].m_LemmaInfo.m_FlexiaModelNo;
//		assert (m_ModelsIndex[debug] <= i);
//		assert (i < m_ModelsIndex[debug + 1]);
//	}
}

bool CMorphDict::Load(wstring wsGrammarFileName)
{
	//      morph.forms_autom      
	//  morph.forms_autom       
	if (!m_pFormAutomat->Load(MakeFName(wsGrammarFileName, L"forms_autom")))
		return false;

	//   morph.annot,    ,  ,
	//  ,   ,  NPS (  )
	wstring wsPrecompiledFile = MakeFName(wsGrammarFileName, L"annot");
	FILE * fp = _wfopen(wsPrecompiledFile.c_str(), L"rb");
	if (!fp)
	{
		string sPrecFile;
		ConvertWStrToStrSysLoc(wsPrecompiledFile, sPrecFile);
		ErrorMessage (Format("Cannot open %s", sPrecFile.c_str()));
		return false;
	}
	
	//         m_FlexiaModels
	ReadFlexiaModels(fp, m_FlexiaModels);

	//         m_AccentModels
	ReadAccentModels(fp, m_AccentModels);
	
	int Count;
	char buffer[256];

	if (!fgets(buffer, 256, fp)) return false;
	Count = atoi(buffer);	//   

	// add empty prefix
	m_Prefixes.resize(1, "");

	//         m_Prefixes
	for (size_t i = 0; i < Count; i++)
	{
		char buffer[256];
		if (!fgets(buffer, 256, fp)) return false;
		string q = buffer;
		Trim(q);
		assert (!q.empty());
		m_Prefixes.push_back(q);
	}

	if (!fgets(buffer, 256, fp)) return false;
	Count = atoi(buffer);	//      

	m_LemmaInfos.clear();

	//            m_LemmaInfos
	ReadVectorInner(fp, m_LemmaInfos, Count);

//	if (!fgets(buffer, 256, fp)) return false;
//	Count = atoi(buffer);	//   NPS    ( )
//
//	m_NPSs.clear();
//	ReadVectorInner(fp, m_NPSs, Count); //   NPS      m_NPSs
//	assert (m_NPSs.size()  == m_FlexiaModels.size());

	fclose(fp);	//   morph.annot

	//     morph.bases      m_Bases
	m_Bases.ReadShortStringHolder(MakeFName(wsGrammarFileName, L"bases"));

	//   (          )
	CreateModelsIndex(); 

	return true;
}

const size_t MaxMrdLineLength = 10240;	//    
void ReadFlexiaModels(FILE* fp, vector<CFlexiaModel>& FlexiaModels )
{
	char buffer[MaxMrdLineLength];	
	if( !fgets(buffer,  MaxMrdLineLength, fp) )
		throw CExpc("Cannot parse mrd file");

	FlexiaModels.clear();

	int paradigm_count = atoi(buffer);	//   

	for(int num = 0; num < paradigm_count; num++)
	{
		if( !fgets(buffer,  MaxMrdLineLength, fp) )	
			throw CExpc("Too few lines in mrd file");

		string line  = buffer;
		Trim(line);

		CFlexiaModel M;

		//          CFlexiaModel
		if (!M.ReadFromString(line))
			throw CExpc(Format("Cannot parse paradigm No %i", num + 1));
		
		FlexiaModels.push_back(M);	//    
	}
}

void ReadAccentModels (FILE* fp, vector<CAccentModel>&	AccentModels)
{
	AccentModels.clear();

	char buffer[MaxMrdLineLength];	
 
	if( !fgets(buffer,  MaxMrdLineLength, fp) )
		throw CExpc("Cannot read accent models from mrd file");
	
	int count = atoi(buffer);	//   
	for(int b = 0; b < count; b++)
	{
		if( !fgets(buffer,  MaxMrdLineLength, fp) )	
			throw CExpc("Too few lines in mrd file");

//		//          CAccentModel
//		//        
//		CAccentModel M;
//
//		if (!M.ReadFromString(buffer))
//			throw CExpc(Format ("Cannot parse line %s", buffer));
//
//		AccentModels.push_back(M);
	}
}

//==============================================================================
const char FlexModelCommDelim[] = "q//q";	//      
bool CFlexiaModel::ReadFromString(string& s)
{
	//  
	int comm = s.rfind(FlexModelCommDelim);
	if (comm != string::npos)
	{
		m_Comments = s.substr(comm+strlen(FlexModelCommDelim));
		Trim(m_Comments);
		s.erase(comm);
		Trim(s);
	}
	else
		m_Comments = "";

	//  ,  - ( )
	StringTokenizer Tok(s.c_str(), "%");

	m_Flexia.clear();

	//    ,  , , 
	while (Tok())
	{
		string OneRecord = Tok.val();
        size_t ast = OneRecord.find('*');	//    
		if (ast == string::npos) return false;
		size_t last_ast = OneRecord.find_last_of('*');
		string  Prefix;
		if (last_ast != ast)
			Prefix = OneRecord.substr(last_ast+1);

		//      CMorphForm: Gramcode, Flexia, Prefix (, , )
		// :      (morph.annot  22.11.2006)     
		CMorphForm G(OneRecord.substr(ast+1, last_ast-ast-1), OneRecord.substr(0, ast), Prefix);
		m_Flexia.push_back(G);
	}

	return true;
}

//==============================================================================
bool		CAccentModel::ReadFromString(const string& s)
{
	//   .     
	// ';', ' '  '\r\n'
	StringTokenizer Tok(s.c_str(), "; \r\n");
	m_Accents.clear();

	//           
	while (Tok())
	{
		string OneRecord = Tok.val();
		if (OneRecord.empty()) return false;
		if (!isdigit(OneRecord[0])) return false;
		m_Accents.push_back(atoi(OneRecord.c_str()));
	}

	return true;
}




